#include <unistd.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "cffread.h"
#include "cfftype.h"
#include "cffutil.h"
#include "unicode_locale.h"
#include "dumputil.h"
#include "fileutil.h"

#define PROGRAM_NAME_AND_VERSION "cffdump 0.1"

/* 할당 테이블도 덤프하려면 아래 매크로를 주석 처리 */
#define SKIP_PRINT_ALLOC_TABLES
/* 스트림을 선택해서 보려면 아래 매크로를 주석 처리 */
#define SKIP_INTERACTIVE_DUMP

#define skip(here)    printf("(skipped: "here")\n")

#define IS_OPTION(x) ((option & (x)) == (x))
static int option = 0;
enum _cffdump_option {
    DESTINATION_DIRECTORY = 1,
    STANDARD_OUTPUT = 1 << 1,
    OVERWRITE = 1 << 2,
    SPLIT = 1 << 3
};
static const char *destination_directory = "";
static const char *source_stream = "";

void    cff_print_header (COMPOUNDFILE *cfp);
void    cff_print_msat   (COMPOUNDFILE *cfp);
void    cff_print_sat    (COMPOUNDFILE *cfp);
void    cff_print_ssat   (COMPOUNDFILE *cfp);
void    cff_print_dirent_struct (CFFDIRENT *dir, DirID_t id);
void    cff_print_all_dirent(COMPOUNDFILE *cfp);
void    cff_print_stream (CFFSTREAM *stream);
void    cff_print_guid   (GUID *guid);
void    cff_print_all    (COMPOUNDFILE *cfp);

void cff_print_dir_list(LNode *base, FILE *fp);
void cff_print_storage_child(int indent,
                             CffDirLNode *parent,
                             LNode *base);
void cff_print_dirent(DirID_t id, LNode *base, COMPOUNDFILE *cfp);

CFFSTREAM *choose_stream (COMPOUNDFILE *cfp, LNode *base);

/* 디렉토리 항목 이름(스트림, 스토리지 이름) 중에
   \005HwpSummaryInformation처럼 앞에 ENQ 문자가
   붙은 것이 있다.  그런 것은 출력되지 않지만 이름 밑에
   표시되는 문자열 길이(널 포함, 바이트 단위)를 보면
   한 글자가 표시되지 않았다는 걸 알 수 있다.
   빈 디렉토리는 이름 문자열이 널 문자 하나로 돼 있고,
   길이는 0으로 저장돼 있다.  */
void cff_print_dirent_name(WCHAR *dirent_name);

/* 윈도우즈, 1601년 기준 100나노초 단위 64비트 시간을
   C언어, 1970년 기준 1초 단위 32비트 시간으로 바꾼다.
   wtime이 0이거나 1970년보다 앞의 시간은 0으로 처리되고
   결과적으로 빈칸으로 출력된다.  */
time_t  wtime_to_ltime  (FILETIME wtime);
void    cff_str_wtime   (char *s, size_t max, FILETIME wtime);
void    cff_print_wtime (FILETIME wtime);

void cff_print_wtime(FILETIME wtime)
{
    char s[64];
    cff_str_wtime(s, 64, wtime);
    printf("%s", s);
}

void cff_str_wtime(char *s, size_t max, FILETIME wtime)
{
    time_t ltime = wtime_to_ltime(wtime);
    if (ltime == 0) {
        s[0] = '\0';
        return;
    }
    strftime(s, max, "%Y-%m-%d %H:%M:%S %Z", localtime(&ltime));
}

time_t wtime_to_ltime(FILETIME wtime)
{
    time_t ltime;
    long long int years = 1970 - 1601;
    long long int leap_years = years/4 - years/100 + years/400;
    long long int diff = (years * 365 + leap_years) * 24 * 60 * 60;

    wtime /= 10000000;
    if (wtime <= diff)
        return 0;

    ltime = wtime - diff;
    return ltime;
}


const static char *dirent_type_str[6] = {
    "(empty)",
    "Storage",
    "Stream",
    "(error)",
    "(error)",
    "Root"
};

static const char *get_dirent_type_str(BYTE type)
{
    if (type >= 0 && type < 6)
        return dirent_type_str[type];
    return "(error)";
}

void cff_print_guid(GUID *guid)
{
    int i;
    if (guid == NULL)
        return;
    printf("%08X-%04X-%04X-%02X%02X-",
           guid->data1, guid->data2, guid->data3,
           guid->data4[0], guid->data4[1]);
    for (i = 2; i < 8; i++)
        printf("%02X", guid->data4[i]);
}


void cff_print_header(COMPOUNDFILE *cfp)
{
    printf("\nCompound File Header (%d bytes):\n", sizeof(CFFHEADER));
    printf("    Signature (8 bytes): ");
    print_byte_seq(cfp->header.signature, 8, PLAIN_MODE);
    printf("    CLSID (16 bytes): ");
    cff_print_guid(&(cfp->header.clsid));
    putchar('\n');

    printf("    Minor Version (2 bytes): %#x\n", cfp->header.minor_version);
    printf("    Major Version (2 bytes): %#x\n", cfp->header.major_version);
    printf("    Byte Order (2 bytes): %#x\n", cfp->header.byte_order);
    printf("    Sector Shift (2 bytes): %d\n", cfp->header.sector_shift);
    printf("    Short Sector Shift (2 bytes): %d\n", cfp->header.short_sector_shift);

    printf("    Reserved (6 bytes): ");
    print_byte_seq(cfp->header.reserved, 6, PLAIN_MODE);

    printf("    Number of Directory Sectors (4 bytes): %d\n", cfp->header.dir_sectors_num);
    printf("    Number of SAT Sectors (4 bytes): %d\n", cfp->header.sat_sectors_num);
    printf("    First SectID of Directory (4 bytes): %d\n", cfp->header.dir_first_sectid);
    printf("    Transaction Signature Number (4 bytes): %d\n", cfp->header.transaction_signature_num);
    printf("    Minimum size of normal stream (4 bytes): %d\n", cfp->header.normal_stream_min_size);
    printf("    First SectID of SSAT (4 bytes): %d\n", cfp->header.ssat_first_sectid);
    printf("    Number of SSAT Sectors (4 bytes): %d\n", cfp->header.ssat_sectors_num);
    printf("    First SectID of MSAT (4 bytes): %d\n", cfp->header.msat_first_sectid);
    printf("    Number of MSAT Sectors (4 bytes): %d\n", cfp->header.msat_sectors_num);
    printf("    First Part of MSAT (4 bytes * 109 items):\n");

#if defined(SKIP_PRINT_ALLOC_TABLES)
    skip("First Part of MSAT");
#else
    print_byte_seq(cfp->header.msat_first_part,
                   sizeof(cfp->header.msat_first_part), NUMBERING_MODE);
#endif

    putchar('\n');
}


void cff_print_msat(COMPOUNDFILE *cfp)
{
    unsigned int total = MSAT_SIZE(cfp);

    printf("\nMaster Sector Allocation Table (using: %d items, "
           "total: %d items):\n", cfp->header.sat_sectors_num,
           total / sizeof(SectID_t));

    print_byte_seq(cfp->msat, total, NUMBERING_MODE);
    putchar('\n');
}


void cff_print_sat(COMPOUNDFILE *cfp)
{
    unsigned int total = OBJ_SIZE(sat, cfp);
    printf("\nSector Allocation Table (%d bytes):\n", total);
    print_byte_seq(cfp->sat, total, NUMBERING_MODE);
    putchar('\n');
}


void cff_print_ssat(COMPOUNDFILE *cfp)
{
    unsigned int total = OBJ_SIZE(ssat, cfp);
    printf("\nShort-Sector Allocation Table (%d bytes): \n", total);
    print_byte_seq(cfp->ssat, total, NUMBERING_MODE);
    putchar('\n');
}


void cff_print_dir_list(LNode *base, FILE *fp)
{
    LNode *n;
    CffDirLNode *ent;

    if (base == NULL)
        return;

    fputs("\n DirID  ParentID   Type      Path\n", fp);
    for (n = base->next; n != NULL; n = n->next) {
        ent = n->data;
        if (ent == NULL)
            break;
        fprintf(fp, " %4d    %5d  %8s   %s\n", ent->id, ent->parent_id,
                get_dirent_type_str(ent->type),
                (ent->path == NULL) ? "" : ent->path);
    }
    fputs("\n", fp);
}


void cff_print_dirent_name(WCHAR *dirent_name)
{
    char name[128];
    if (dirent_name == NULL)
        return;
    ul_utf16_to_locale(dirent_name, 32, name, 128);
    printf("%s", name);
}

void cff_print_dirent_struct(CFFDIRENT *dir, DirID_t id)
{
    if (dir == NULL)
        return;

    printf("\nDirID: %d\n", id);
    printf("    Name (64 bytes): ");
    cff_print_dirent_name(dir[id].name);
    putchar('\n');

    printf("    Name Length (2 bytes): %d\n", dir[id].name_len);
    printf("    Object Type (1 bytes): %s\n",
           get_dirent_type_str(dir[id].object_type));
    printf("    Color Flag (1 bytes): %#x\n", dir[id].color_flag);
    printf("    Left DirID (4 bytes): %d\n", dir[id].left_dirid);
    printf("    Right DirID (4 bytes): %d\n", dir[id].right_dirid);
    printf("    Child DirID (4 bytes): %d\n", dir[id].child_dirid);
    printf("    CLSID (16 bytes): ");
    cff_print_guid(&(dir[id].clsid));
    putchar('\n');

    printf("    State Bits (4 bytes): %#x\n", dir[id].state_bits);
    printf("    Creation Time (8 bytes): ");
    cff_print_wtime(dir[id].creation_time);
    putchar('\n');
    printf("    Modified Time (8 bytes): ");
    cff_print_wtime(dir[id].modified_time);
    putchar('\n');

    printf("    First SectID of Object (4 bytes): %d\n", dir[id].first_sectid);
#ifdef IN_WINDOWS
    printf("    Stream Object Size (8 bytes): %I64d\n", dir[id].stream_size);
#else
    printf("    Stream Object Size (8 bytes): %lld\n", dir[id].stream_size);
#endif
}


void cff_print_all_dirent(COMPOUNDFILE *cfp)
{
    int i;
    unsigned int total;

    LNode *base = cff_make_dir_list(cfp);
    if (base == NULL)
        return;
    CffDirLNode *root = base->next->data;

    printf("\nDirectory structure:\n");
    printf("%s\n", root->name);
    cff_print_storage_child(4, root, base);
    putchar('\n');

    printf("\nDirectory table:\n");
    cff_print_dir_list(base, stdout);

    total = CFF_DIR_ENTRY_NUM(cfp);
    printf("\n\nDirectory Stream (%d entries):\n", total);
    for (i = 0; i < total; i++)
        cff_print_dirent(i, base, cfp);

    putchar('\n');
    cff_destroy_dir_list(base);
}


void cff_print_storage_child(int indent,
                             CffDirLNode *parent,
                             LNode *base)
{
    LNode *n;
    CffDirLNode *cn;
    int i;

    if (parent == NULL || base == NULL)
        return;

    for (n = base->next; n != NULL; n = n->next) {
        cn = n->data;
        if (cn == NULL)
            return;
        if (parent->id == cn->parent_id) {
            for (i = 0; i < indent; ++i)
                putchar(' ');
            printf("%s\n", (cn->name == NULL) ? "" : cn->name);
            if (cn->type == CFF_STORAGE_DIR || cn->type == CFF_ROOT_DIR)
                cff_print_storage_child(indent + 4, cn, base);
        }
    }
}

void cff_print_dirent(DirID_t id, LNode *base, COMPOUNDFILE *cfp)
{
    CffDirLNode *ent;
    CFFSTREAM *stream;

    if (base == NULL)
        return;

    ent = cff_find_dir_node_by_id(id, base);
    if (ent == NULL)
        return;

    cff_print_dirent_struct(cfp->dir, id);
    putchar('\n');

    switch (cfp->dir[id].object_type) {
    case CFF_STREAM_DIR:
        stream = cff_salloc(ent->path, cfp);
        if (stream == NULL)
            return;
        cff_print_stream(stream);
        cff_free(stream);
        break;
    case CFF_ROOT_DIR:
        stream = cff_salloc_by_id(0, cfp);
        if (stream == NULL)
            return;
        cff_print_stream(stream);
        cff_free(stream);
        /* 루트는 스트림(sscs)이면서 스토리지이다. fall through */
    case CFF_STORAGE_DIR:
#ifndef SKIP_INTERACTIVE_DUMP
        printf("Directory structure:\n");
        printf("%s\n", (ent->name == NULL) ? "" : ent->name);
        cff_print_storage_child(4, ent, base);
#endif
        break;
    default:
        break;
    }
    putchar('\n');
}

void cff_print_stream(CFFSTREAM *stream)
{
    if (stream == NULL)
        return;
    printf("Stream byte sequence (%d bytes):\n", stream->data_size);
    print_byte_seq(stream->data, stream->data_size, NUMBERING_MODE);
    putchar('\n');
}


void cff_choose_print_dirent(COMPOUNDFILE *cfp)
{
    unsigned int id;
    char dirid_str[8];
    LNode *base = cff_make_dir_list(cfp);
    if (base == NULL)
        return;

    while (1) {
        fputs("\nDirectory list:\n", stderr);
        cff_print_dir_list(base, stderr);
        fputs("Input DirID (q=quit) : ", stderr);
        fgets(dirid_str, 8, stdin);
        if (dirid_str[0] == 'q')
            break;
        if (1 != sscanf(dirid_str, "%u", &id))
            continue;
        cff_print_dirent(id, base, cfp);
    }

    cff_destroy_dir_list(base);
}

void cff_print_all(COMPOUNDFILE *cfp)
{
    if (cfp == NULL)
        return;

    cff_print_header(cfp);

#if defined(SKIP_PRINT_ALLOC_TABLES)
    skip("MSAT, SAT, SSAT");
#else
    cff_print_msat(cfp);
    cff_print_sat(cfp);
    cff_print_ssat(cfp);
#endif

#if defined(SKIP_INTERACTIVE_DUMP)
    cff_print_all_dirent(cfp);
#else
    cff_choose_print_dirent(cfp);
#endif
}


static void help_msg()
{
    printf("Usage: cffdump [OPTION] FILE...\n"
           "Dump compound file format FILEs\n\n"
           "Option:\n"
           "  -p              Print to standard output\n"
           "  -d DIRECTORY    Create output files in DIRECTORY\n"
           "  -o              Overwrite output files\n"
           "  -s              Split FILE into normal files\n"
           "  -ss STREAM      Split STREAM or storage of FILE into normal files\n"
           "  --              Next arguments are not an option\n"
           "  --help\n"
           "  --version\n\n"
           "Report bugs to <http://kldp.net/projects/cffdump/>\n");
#ifdef IN_WINDOWS
    printf("\nIn Windows, you can drag and drop cff files "
           "on the program icon\n");
    press_any_key();
#endif
    exit(0);
}

static void version_msg()
{
    printf(PROGRAM_NAME_AND_VERSION"\n"
           "Written by Young ho Park\n"
           "This program is following Open Specification Promise of Microsoft\n");
#ifdef IN_WINDOWS
    press_any_key();
#endif
    exit(0);
}

/* path를 이용해 새 파일 이름을 만들고 표준출력을 파일로 돌린다.
   성공시 0, 실패시 -1 반환한다.  */
static int redirect_cffdump_file(const char *path)
{
    return redirect_stdout_to_file(!IS_OPTION(OVERWRITE),
                                   path, "cffdump.txt");
}

static int print_cff(const char *path, COMPOUNDFILE *cfp)
{
    if (!IS_OPTION(STANDARD_OUTPUT))
        if (-1 == redirect_cffdump_file(path))
            return -1;

    printf("'%s'\n", path);
    cff_print_all(cfp);
    return 0;
}

static CffDirLNode *get_source_stream_node(LNode *base)
{
    if (base == NULL)
        return NULL;
    if (source_stream[0] == '\0')
        return base->next->data;

    return cff_find_dir_node(source_stream, base);
}

static int split_cff(const char *path, COMPOUNDFILE *cfp)
{
    CffDirLNode *start_stream;
    LNode *list;

    list = cff_make_dir_list(cfp);
    if (list == NULL)
        return -1;

    start_stream = get_source_stream_node(list);
    if (start_stream == NULL) {
        fprintf(stderr, "Cannot find %s in %s\n", source_stream, path);
        goto error;
    }

    /* 스트림 이름이 빈문자열인 경우도 있을까   */
    if (start_stream->name[0] == '\0') {
        char *str = start_stream->name;
        free(str);
        start_stream->name = strdup("none");
        if (start_stream->name == NULL)
            goto error;
    }

    if (-1 == make_dir_chdir_with_suffix(path, "cff_split"))
        return -1;

    cff_make_split_stream_storage(start_stream, list, cfp,
                                  NULL, !IS_OPTION(OVERWRITE), "");
    cff_destroy_dir_list(list);
    return 0;

error:
    cff_destroy_dir_list(list);
    return -1;
}

static int cff_file_dump(const char *path)
{
    COMPOUNDFILE *cfp;

    if (path == NULL)
        return -1;

    cfp = cff_open(path, "rb");
    if (cfp == NULL) {
        fprintf(stderr, "Cannot open %s\n", path);
        return 1;
    }

    if (IS_OPTION(DESTINATION_DIRECTORY))
        if (-1 == make_dir_chdir(destination_directory)) {
            cff_close(cfp);
            return -1;
        }

    if (IS_OPTION(SPLIT))
        split_cff(path, cfp);
    else
        print_cff(path, cfp);

    cff_close(cfp);
    return 0;
}

static void unknown_option_exit(char opt)
{
    fprintf(stderr, "Unknown option: %c\n", opt);
    exit(1);
}

static void help_option_arg_usage(char opt)
{
    switch(opt) {
    case 'd':
        fputs("Usage: cffdump -d DIRECTORY FILE...\n", stderr);
        break;
    case 's':
        fputs("Usage: cffdump -ss STREAM FILE...\n", stderr);
        break;
    case '-':
        fputs("Usage: cffdump -- FILE...\n", stderr);
        break;
    default:
        unknown_option_exit(opt);
        break;
    }
    exit(1);
}

static void read_option_arg(char opt, int *i, int argc, char **argv)
{
    *i += 1;
    if (*i >= argc)
        help_option_arg_usage(opt);

    switch (opt) {
    case 'd':
        destination_directory = argv[*i];
        break;
    case 's':
        source_stream = argv[*i];
        break;
    case '-':
        break;
    default:
        unknown_option_exit(opt);
        break;
    }
}

static void read_option(int *file_start, int argc, char **argv)
{
    int i;
    char opt;

    for (i = 1; i < argc; ++i) {
        if (argv[i][0] == '-') {
            opt = argv[i][1];
            switch (opt) {
            case 'd':
                option |= DESTINATION_DIRECTORY;
                read_option_arg(opt, &i, argc, argv);
                break;
            case 'p':
                option |= STANDARD_OUTPUT;
                break;
            case 'o':
                option |= OVERWRITE;
                break;
            case 's':
                option |= SPLIT;
                if (argv[i][2] == opt)
                    read_option_arg(opt, &i, argc, argv);
                break;
            case '-':
                read_option_arg(opt, &i, argc, NULL);
                goto option_finish;
            default:
                unknown_option_exit(opt);
                break;
            }
        } else {
            break;
        }
    }

option_finish:
    *file_start = i;
    if (*file_start < argc)
        return;
    fputs("Usage: cffdump [OPTION] FILE...\n", stderr);
    exit(1);
}

int main(int argc, char **argv)
{
    int i;
    int err;
    int file_start = 1;
    char *start_dir;

    if (argc <= 1 || 0 == strcmp("--help", argv[1]))
        help_msg();
    else if (0 == strcmp("--version", argv[1]))
        version_msg();

    read_option(&file_start, argc, argv);

    start_dir = strdup_pwd();
    if (start_dir == NULL)
        return 1;

    for (i = file_start; i < argc; ++i) {
        if (-1 == cff_file_dump(argv[i]))
            break;
        if (-1 == chdir(start_dir)) {
            err = errno;
            file_err_msg(err, start_dir);
            break;
        }
    }

    free(start_dir);
#ifdef IN_WINDOWS
    press_any_key();
#endif

    return 0;
}
