#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "cffread.h"
#include "cfftype.h"
#include "unicode_locale.h"

#define CFP_SEEK_SECTOR(id, cfp) \
    fseek((cfp)->fp, (cfp)->sector_size * (1 + (id)), SEEK_SET)
#define PTR_READ_SECTOR(ptr, cfp) \
    fread((ptr), (cfp)->sector_size, 1, (cfp)->fp)

#define P_ADD(ptr, nbytes)  ((BYTE *) ptr + (nbytes))

/* sat를 이용해서 스트림을 읽는다.
   dest: 저장할 포인터
   id: 스트림의 섹터 시작 번호
   nsect: 스트림의 섹터 개수
   cfp: compound file pointer  */
static void read_stream_in_sat(void *dest,
                               SectID_t id,
                               unsigned nsect,
                               COMPOUNDFILE *cfp)
{
    unsigned int i;
    for (i = 0; i < nsect; i++) {
        CFP_SEEK_SECTOR(id, cfp);
        PTR_READ_SECTOR(dest, cfp);
        id = cfp->sat[id];
        dest = P_ADD(dest, cfp->sector_size);
        if (id == CFF_EOC_SECTID)
            break;
    }
}

/* ssat를 이용해서 short 스트림을 읽는다.  사용법은 위와 같다.  */
static void read_stream_in_ssat(void *dest,
                                SectID_t id,
                                unsigned nsect,
                                COMPOUNDFILE *cfp)
{
    unsigned int i;
    BYTE *src;
    for (i = 0; i < nsect; ++i) {
        src = (BYTE *) cfp->sscs->data;
        src = P_ADD(src, id * cfp->short_sector_size);
        memcpy(dest, src, cfp->short_sector_size);
        dest = P_ADD(dest, cfp->short_sector_size);
        id = cfp->ssat[id];
        if (id == CFF_EOC_SECTID)
            break;
    }
}

static int  get_header (COMPOUNDFILE  *cfp);

/* 여기서 at는 Allocation Table의 약자: msat, sat, ssat */
static int  alloc_msat (COMPOUNDFILE  *cfp);
static int  alloc_sat  (COMPOUNDFILE  *cfp);
static int  alloc_ssat (COMPOUNDFILE  *cfp);
static void free_at    (SectID_t **at);

static int  alloc_dir  (COMPOUNDFILE  *cfp);
static void free_dir   (CFFDIRENT **dir);

static DirID_t  get_dirid_by_name (const utf16char *name,
                                   DirID_t node_id,
                                   const CFFDIRENT *d);

static int      alloc_sscs (COMPOUNDFILE *cfp);


/* 헤더를 읽는다.  성공시 0, 실패시 -1을 반환한다.  */
static int get_header(COMPOUNDFILE *cfp)
{
    int c;
    const static BYTE sign[8] = CFF_SIGNATURE;

    fread(&(cfp->header), sizeof(CFFHEADER), 1, cfp->fp);
    if (ferror(cfp->fp))
        goto error_handler;

    c = memcmp(sign, cfp->header.signature, sizeof(sign));
    if (0 != c)
        goto error_handler;

    return 0;

error_handler:
    memset(&(cfp->header), 0, sizeof(CFFHEADER));
    return -1;
}

/* 헤더 밖에 있는 msat를 읽는다.  성공 0, 실패 -1 */
static int append_msat_remainder(COMPOUNDFILE *cfp)
{
    DWORD copy_size;  /* 한 섹터에서 실제 복사하는 크기 */
    DWORD next_index; /* 다음 섹터 번호가 있는 곳 */
    DWORD i;
    SectID_t id;    /* 다음 섹터 번호 */
    SectID_t *temp; /* 임시 섹터 */
    BYTE *ptr = (BYTE *) cfp->msat;

    temp = malloc(cfp->sector_size);
    if (temp == NULL)
        goto error_handler;

    ptr = P_ADD(ptr, CFF_HEADER_MSAT_SIZE);
    id = cfp->header.msat_first_sectid;

    copy_size = cfp->sector_size - sizeof(SectID_t);
    next_index = copy_size / sizeof(SectID_t);

    /* 파일이 정상이면 개수만큼만 읽어도 되지만
       정상이 아닐 수도 있으니 루프 마지막에서 다음 id가
       EOC(End Of Chain)인지 확인해주었다.
       다른 부분도 이런 식으로 확인했다.  */
    for (i = 0; i < NSECTORS(msat, cfp); i++) {
        CFP_SEEK_SECTOR(id, cfp);
        PTR_READ_SECTOR(temp, cfp);
        memcpy(ptr, temp, copy_size);
        ptr = P_ADD(ptr, copy_size);
        id = temp[next_index];
        if (id == CFF_EOC_SECTID)
            break;
    }

    free(temp);
    if (ferror(cfp->fp))
        goto error_handler;

    return 0;

error_handler:
    free_at(&(cfp->msat));
    return -1;
}


/* msat를 할당한다. 성공 0, 실패 -1 */
static int alloc_msat(COMPOUNDFILE *cfp)
{
    cfp->msat = malloc(MSAT_SIZE(cfp));
    if (cfp->msat == NULL)
        return -1;

    memcpy(cfp->msat, cfp->header.msat_first_part, CFF_HEADER_MSAT_SIZE);

    if (NSECTORS(msat, cfp) > 0)
        if (-1 == append_msat_remainder(cfp))
            return -1;

    return 0;
}

/* Allocation Table 메모리 해제 */
static void free_at(SectID_t **at)
{
    if (at == NULL)
        return;
    free(*at);
    *at = NULL;
}

/* sat 할당. 성공 0, 실패 -1 */
static int alloc_sat(COMPOUNDFILE *cfp)
{
    DWORD i;
    BYTE *ptr;

    cfp->sat = malloc(OBJ_SIZE(sat, cfp));
    ptr = (BYTE *) cfp->sat;
    if (ptr == NULL)
        return -1;

    for (i = 0; i < NSECTORS(sat, cfp); i++) {
        CFP_SEEK_SECTOR(cfp->msat[i], cfp);
        PTR_READ_SECTOR(ptr, cfp);
        ptr = P_ADD(ptr, cfp->sector_size);
        if (cfp->msat[i + 1] == CFF_FREE_SECTID)
            break;
    }

    if (ferror(cfp->fp)) {
        free_at(&(cfp->sat));
        return -1;
    }

    return 0;
}

/* ssat 할당. 성공시 0, 실패시 -1, 없다면 1을 반환한다.
   모든 섹터가 일반이면 short sector가 없을 수도 있다.  */
static int alloc_ssat(COMPOUNDFILE *cfp)
{
    SectID_t id = cfp->header.ssat_first_sectid;
    if (id == CFF_EOC_SECTID || OBJ_SIZE(ssat, cfp) == 0) {
        /* 루트 스트림에 sscs이 없는지 확인 */
        if (cfp->dir[0].stream_size == 0)
            return 1;
        else
            return -1;
    }

    cfp->ssat = malloc(OBJ_SIZE(ssat, cfp));
    if (cfp->ssat == NULL)
        return -1;

    read_stream_in_sat(cfp->ssat, id, NSECTORS(ssat, cfp), cfp);

    if (ferror(cfp->fp)) {
        free_at(&(cfp->ssat));
        return -1;
    }
    return 0;
}

/* Compound File Format의 버전이 3이면
   디렉토리 스트림 섹터 개수가 파일에 없다.  */
static int save_dir_sectors_num(SectID_t id, COMPOUNDFILE *cfp)
{
    unsigned int cnt = 0;
    while (id != CFF_EOC_SECTID) {
        id = cfp->sat[id];
        ++cnt;
        if (cnt > CFF_MAX_REG_SECTID)
            return -1;
    }
    cfp->header.dir_sectors_num = cnt;
    return 0;
}

/* directory 스트림 할당.  성공 0, 실패 -1  */
static int alloc_dir(COMPOUNDFILE *cfp)
{
    SectID_t id = cfp->header.dir_first_sectid;
    if (id == CFF_EOC_SECTID)
        return -1;

    if (cfp->header.major_version < CFF_MAJOR_VER_4
        || NSECTORS(dir, cfp) == 0)
        if (-1 == save_dir_sectors_num(id, cfp))
            return -1;

    cfp->dir = malloc(OBJ_SIZE(dir, cfp));
    if (cfp->dir == NULL)
        return -1;

    read_stream_in_sat(cfp->dir, id, NSECTORS(dir, cfp), cfp);

    if (ferror(cfp->fp)) {
        free_dir(&(cfp->dir));
        return -1;
    }

    return 0;
}

/* dir 메모리 해제 */
static void free_dir(CFFDIRENT **dir)
{
    if (dir == NULL)
        return;
    free(*dir);
    *dir = NULL;
}

/* directory 이름으로 나무에서 id를 찾는다.
   d는 directory 동적 배열이다.  */
static DirID_t get_dirid_by_name(const utf16char *name,
                                 DirID_t node_id,
                                 const CFFDIRENT *d)
{
    DirID_t id;

    if (CFF_NO_DIRID == node_id)
        return CFF_NO_DIRID;

    if (0 == memcmp(name, d[node_id].name, d[node_id].name_len))
        return node_id;

    id = get_dirid_by_name(name, d[node_id].left_dirid, d);
    if (id != CFF_NO_DIRID)
        return id;

    return get_dirid_by_name(name, d[node_id].right_dirid, d);
}

/* cff_get_dirid에서 쓰는 함수. path_copy에서 첫글자를 조사한다.
   맨앞에 슬래시가 있으면 id는 루트 번호로 설정한다.  */
static char *check_first_ch(DirID_t *id, char *path_copy, const COMPOUNDFILE *cfp)
{
    if (*path_copy == '/') {
        *id = 0;
        return path_copy + 1;
    } else {
        *id = cfp->current_dirid;
        return path_copy;
    }
}

/* directory 스트림 경로로 id를 찾아준다.  */
unsigned int cff_get_dirid(const char *path, COMPOUNDFILE *cfp)
{
    DirID_t id;
    char *copy;
    char *chr_ptr;
    char *slash;
    char *next;
    utf16char name[32];

    if (path == NULL)
        return CFF_NO_DIRID;

    copy = strdup(path);
    if (copy == NULL)
        return CFF_NO_DIRID;

    chr_ptr = check_first_ch(&id, copy, cfp);
    if (*chr_ptr == '\0')
        goto return_part;

    while (1) {
        slash = strchr(chr_ptr, '/');
        if (slash != NULL) {
            *slash = '\0';
            next = slash + 1;
        } else {
            next = "";
        }
        ul_locale_to_utf16(chr_ptr, 0, name, 32);
        id = get_dirid_by_name(name,
                               cfp->dir[id].child_dirid,
                               cfp->dir);
        if (*next == '\0')
            break;
        chr_ptr = next;
    }

return_part:
    free(copy);
    return id;
}

#define GET_ALLOC_SIZE(data_size, sector_size) \
    ((((data_size) - 1) / (sector_size) + 1) * (sector_size))
/* 스트림 구조체에 메모리를 할당해서 반환한다.  cff_free로 해제한다.  */
static CFFSTREAM *alloc_stream(size_t data_size, size_t sector_size)
{
    CFFSTREAM *stream;
    if (data_size == 0 || sector_size == 0)
        return NULL;
    stream = malloc(sizeof(CFFSTREAM));
    if (stream == NULL)
        return NULL;
    stream->alloc_size = GET_ALLOC_SIZE(data_size, sector_size);
    stream->data_size = data_size;
    stream->data = malloc(stream->alloc_size);
    if (stream->data == NULL) {
        free(stream);
        return NULL;
    }
    return stream;
}

/* 스트림의 크기와 시작 id를 확인한다.  ok: 0, error: -1  */
static int check_size_sectid(size_t *data_size, SectID_t *id, const CFFDIRENT *d)
{
    *data_size = (size_t) d->stream_size;
    if (*data_size == 0)
        return -1;

    *id = d->first_sectid;
    if (*id == CFF_EOC_SECTID)
        return -1;

    return 0;
}

/* 스트림 크기에 따라 일반과 작은 스트림으로 나뉜다.  */
typedef enum {
    _CFF_STREAM_ERROR,
    _CFF_STREAM_SHORT,
    _CFF_STREAM_NORMAL
} CffStreamType;

static CffStreamType get_stream_type(DirID_t id, COMPOUNDFILE *cfp)
{
    switch (cfp->dir[id].object_type) {
    case CFF_ROOT_DIR:
        return _CFF_STREAM_NORMAL;
    case CFF_STREAM_DIR:
        if (cfp->dir[id].stream_size < cfp->header.normal_stream_min_size)
            return _CFF_STREAM_SHORT;
        else
            return _CFF_STREAM_NORMAL;
    default:
        return _CFF_STREAM_ERROR;
    }
}

/* 디렉토리 스트림 id로 찾은 스트림을 반환한다.  cff_free로 해제한다.  */
CFFSTREAM *cff_salloc_by_id(unsigned int id, COMPOUNDFILE *cfp)
{
    SectID_t sectid;
    CFFSTREAM *stream;
    size_t data_size;
    CffStreamType type;
    unsigned int nsect;

    if (id == CFF_NO_DIRID)
        return NULL;
    if (-1 == check_size_sectid(&data_size, &sectid, &(cfp->dir[id])))
        return NULL;

    type = get_stream_type(id, cfp);
    switch (type) {
    case _CFF_STREAM_NORMAL:
        stream = alloc_stream(data_size, cfp->sector_size);
        if (stream == NULL)
            return NULL;
        nsect = stream->alloc_size / cfp->sector_size;
        read_stream_in_sat(stream->data, sectid, nsect, cfp);
        if (ferror(cfp->fp)) {
            cff_free(stream);
            return NULL;
        }
        break;
    case _CFF_STREAM_SHORT:
        if (cfp->sscs == NULL)
            return NULL;
        stream = alloc_stream(data_size, cfp->short_sector_size);
        if (stream == NULL)
            return NULL;
        nsect = stream->alloc_size / cfp->short_sector_size;
        read_stream_in_ssat(stream->data, sectid, nsect, cfp);
        break;
    default: /* error case */
        return NULL;
    }

    return stream;
}

/* short sector container stream을 할당한다.
   short stream이 없으면 sscs도 없으므로 없다고
   에러는 아니다.  없으면 1, 있으면 0 반환한다. */
static int alloc_sscs(COMPOUNDFILE *cfp)
{
    cfp->sscs = cff_salloc_by_id(0, cfp);
    if (NULL == cfp->sscs)
        return 1;
    return 0;
}

/* 경로에 해당하는 스트림을 반환한다.  */
CFFSTREAM *cff_salloc(const char *stream_path, COMPOUNDFILE *cfp)
{
    CFFSTREAM *stream;
    DirID_t id = cff_get_dirid(stream_path, cfp);
    if (id == CFF_NO_DIRID)
        return NULL;
    if (cfp->dir[id].object_type != CFF_STREAM_DIR)
        return NULL;

    stream = cff_salloc_by_id(id, cfp);
    if (stream == NULL)
        return NULL;

    return stream;
}


void cff_free(CFFSTREAM *stream)
{
    if (stream == NULL)
        return;
    free(stream->data);
    free(stream);
}


COMPOUNDFILE *cff_open(const char *path, const char *mode)
{
    COMPOUNDFILE *cfp = malloc(sizeof(COMPOUNDFILE));
    if (cfp == NULL)
        return NULL;

    memset(cfp, 0, sizeof(COMPOUNDFILE));
    cfp->fp = fopen(path, mode);

    if (cfp->fp == NULL)
        goto error_handler;

    if (get_header(cfp) == -1)
        goto error_handler;

    cfp->sector_size = 1 << cfp->header.sector_shift;
    cfp->short_sector_size = 1 << cfp->header.short_sector_shift;

    if (alloc_msat(cfp) == -1)
        goto error_handler;

    if (alloc_sat(cfp) == -1)
        goto error_handler;

    if (alloc_dir(cfp) == -1)
        goto error_handler;

    if (alloc_ssat(cfp) == -1)
        goto error_handler;

    /* sscs와 ssat는 한배를 탄다.
       나머지 경우는 alloc_ssat에서 처리했다.  */
    if (alloc_sscs(cfp) == 1 && cfp->ssat != NULL)
        goto error_handler;

    return cfp;

error_handler:
    cff_close(cfp);
    return NULL;
}


void cff_close(COMPOUNDFILE *cfp)
{
    if (cfp == NULL)
        return;
    if (cfp->fp != NULL)
        fclose(cfp->fp);
    free_at(&(cfp->msat));
    free_at(&(cfp->sat));
    free_at(&(cfp->ssat));
    free_dir(&(cfp->dir));
    cff_free(cfp->sscs);
    free(cfp);
}


int cff_change_storage(const char *path, COMPOUNDFILE *cfp)
{
    DirID_t id = cff_get_dirid(path, cfp);
    if (id == CFF_NO_DIRID)
        return -1;

    switch (cfp->dir[id].object_type) {
    case CFF_STORAGE_DIR:
    case CFF_ROOT_DIR:
        cfp->current_dirid = id;
        return 0;
    case CFF_UNALLOCATED_DIR:
    case CFF_STREAM_DIR:
        return -1;
    }

    return -1;
}
