#ifndef CFFTYPE_H
#define CFFTYPE_H

#include <inttypes.h>
#include <stdio.h>
#include "cffread.h"

/* 구조체 정렬 1바이트 */
#pragma pack(1)

/* 타입명은 OpenOffice.org의 Microsoft Compound Document
   File Format 문서에 나오는 용어에서 주로 따왔다.
   MS-CFB 문서의 주요 용어와 비교하면

        섹터의 순서를 저장한 표를 나열하는 표
        MS: Double-Indirect File Allocation Table (DIFAT)
        OO: Master Sector Allocation Table (MSAT)

        섹터의 순서를 저장한 표
        MS: File Allocation Table (FAT)
        OO: Sector Allocation Table (SAT)

        64바이트짜리 섹터들의 순서를 저장한 표
        MS: Mini FAT
        OO: Short SAT

        64바이트짜리 섹터
        MS: Mini Sector
        OO: Short Sector

        64바이트짜리 섹터들의 저장 공간
        MS: Mini Stream
        OO: Short Stream Container Stream

        4096바이트보다 작은 스트림
        MS: Stream (명칭 구분 없다.)
        OO: Short Stream

        디렉토리 항목 위치에 따른 순번
        (하나의 디렉토리 항목은 하나의 스트림에 대응한다.)
        MS: Stream ID
        OO: DirID

        디렉토리 항목 배열 또는 스트림
        MS: Directory Entry Array
        OO: Directory Stream        */


/* 마이크로소프트 윈도우즈 데이터 타입 */

typedef struct {
    uint32_t    data1;
    uint16_t    data2;
    uint16_t    data3;
    uint8_t     data4[8];
} GUID;

typedef uint64_t FILETIME;
typedef uint64_t ULONGLONG;
typedef uint32_t DWORD;
typedef uint16_t USHORT;
typedef uint8_t  BYTE;
typedef uint16_t WCHAR;

/* 복합파일 표시 */
#define CFF_SIGNATURE {0xD0, 0xCF, 0x11, 0xE0, 0xA1, 0xB1, 0x1A, 0xE1}

/* 복합파일 데이터 타입 */

typedef DWORD SectID_t;
#define CFF_HEADER_MSAT_NUM  109
/* 파일이 클 경우 MSAT가 일반 섹터에도 있을 수 있으므로 주의 */
#define CFF_HEADER_MSAT_SIZE (CFF_HEADER_MSAT_NUM * sizeof(SectID_t))

typedef struct {
    BYTE     signature[8];
    GUID     clsid;
    USHORT   minor_version;
    USHORT   major_version;
    USHORT   byte_order;
    USHORT   sector_shift;
    USHORT   short_sector_shift;
    BYTE     reserved[6];
    DWORD    dir_sectors_num;
    DWORD    sat_sectors_num;
    SectID_t dir_first_sectid;
    DWORD    transaction_signature_num;
    DWORD    normal_stream_min_size;
    SectID_t ssat_first_sectid;
    DWORD    ssat_sectors_num;
    SectID_t msat_first_sectid;
    DWORD    msat_sectors_num;
    SectID_t msat_first_part[CFF_HEADER_MSAT_NUM];
} CFFHEADER;

/* 헤더에서 고정된 값들 */
#define CFF_CLSID_NULL     {0, 0, 0, {0,0,0,0,0,0,0,0}}
#define CFF_MINOR_VER      0x003E
#define CFF_MAJOR_VER_3    0x0003
#define CFF_MAJOR_VER_4    0x0004
#define CFF_LITTLE_ENDIAN  0xFFFE

/* 파일의 메이저 버전에 따라 섹터 크기 결정 */
#define CFF_SECTOR_SHIFT_V3  0x0009
#define CFF_SECTOR_SHIFT_V4  0x000C
#define CFF_SECTOR_SIZE_V3   512
#define CFF_SECTOR_SIZE_V4   4096

/* Short 섹터 크기는 버전과 무관하다. */
#define CFF_SHORT_SECTOR_SHIFT  0x0006
#define CFF_SHORT_SECTOR_SIZE   64

/* 헤더의 예비 공간은 0으로 고정해야 함 */
#define CFF_RESERVED  {0,0,0,0,0,0}

/* 복합파일 버전3 헤더에서는 디렉토리 섹터 개수(dir_sectors_num)
   변수를 지원하지 않는다. 그러나 cffread.c는 파일 여는 과정에서
   디렉토리 섹터 개수를 그 변수에 임시로 저장해서 실행 중에 사용한다. */
#define CFF_DIR_SECTORS_NUM_V3  0

/* File Transaction 기능 구현 없으면 0으로 고정 */
#define CFF_NO_TRANSACTION      0

/* 일반 스트림의 최소 크기 */
#define CFF_NORM_STRM_MIN_SIZE  4096

/* 헤더 크기 */
#define CFF_HEADER_SIZE 512

/* 버전3의 섹터0 시작 위치 */
#define CFF_SECTOR0_V3  512

/* 버전4의 섹터0 시작 위치, 이곳과 헤더 사이는 0으로 채움 */
#define CFF_SECTOR0_V4  4096

/* 섹터 할당 표(SAT)에서 사용하는 값 */
#define CFF_FREE_SECTID      0xFFFFFFFF
#define CFF_EOC_SECTID       0xFFFFFFFE
#define CFF_SAT_SECTID       0xFFFFFFFD
#define CFF_MSAT_SECTID      0xFFFFFFFC

/* 일반 섹터 ID의 최대값 */
#define CFF_MAX_REG_SECTID   0xFFFFFFFA

typedef DWORD    DirID_t;

typedef struct {
    WCHAR      name[32];
    USHORT     name_len; /* 널 문자 포함 바이트 단위 */
    BYTE       object_type;
    BYTE       color_flag;
    DirID_t    left_dirid;
    DirID_t    right_dirid;
    DirID_t    child_dirid;
    GUID       clsid;
    DWORD      state_bits;
    FILETIME   creation_time;
    FILETIME   modified_time;
    SectID_t   first_sectid;
    ULONGLONG  stream_size;
} CFFDIRENT;

/* 디렉토리 항목 크기 */
#define CFF_DIRENT_SIZE     128
#define CFF_DIRENT_NUM_PER_SECT_V3  (CFF_SECTOR_SIZE / CFF_DIRENT_SIZE)

/* 디렉토리 항목 구조체의 object_type 값 */
#define CFF_UNALLOCATED_DIR 0x00
#define CFF_STORAGE_DIR     0x01
#define CFF_STREAM_DIR      0x02
#define CFF_ROOT_DIR        0x05

/* 디렉토리 항목 구조체의 color_flag 값 */
#define CFF_RED_FLAG        0x00
#define CFF_BLACK_FLAG      0x01

/* 디렉토리 항목이 없을 때 쓰는 값 */
#define CFF_NO_DIRID        0xFFFFFFFF

struct cff_compoundfile_struct {
    CFFHEADER   header;
    FILE       *fp;
    SectID_t   *msat;
    SectID_t   *sat;
    SectID_t   *ssat;
    CFFDIRENT  *dir;
    CFFSTREAM  *sscs;
    DWORD       sector_size;
    DWORD       short_sector_size;
    DirID_t     current_dirid;
};

/* obj에는 msat, sat, ssat, dir이 올 수 있다.
   cfp에는 COMPOUNDFILE 포인터가 들어간다.  */
#define NSECTORS(obj, cfp) \
    ((cfp)->header.obj##_sectors_num)
/* 아래 obj에는 sat, ssat, dir이 올 수 있다.  */
#define OBJ_SIZE(obj, cfp) \
    (NSECTORS(obj, cfp) * (cfp)->sector_size)
/* msat는 크기를 다르게 계산한다.  */
#define MSAT_SIZE(cfp) \
    (NSECTORS(msat, cfp) * ((cfp)->sector_size - sizeof(SectID_t)) + CFF_HEADER_MSAT_SIZE)

#define CFF_DIR_ENTRY_NUM(cfp) \
    (OBJ_SIZE(dir, cfp) / sizeof(CFFDIRENT))

#endif /* CFFTYPE_H */
