#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "fileutil.h"
#include "unicode_string.h"
#include "cffread.h"
#include "cfftype.h"
#include "cffutil.h"
#include "linkedlist.h"

static void collect_sibling_name(LNode *b, DirID_t id, COMPOUNDFILE *cfp);
static LNode *make_direct_child_name_list(const char *storage_path,
                                          COMPOUNDFILE *cfp);

static void free_entry(CffDirLNode *ent);
static char *make_new_path(const char *prefix, const char *append);

static void add_path_parent(CffDirLNode *parent,
                            LNode *base,
                            COMPOUNDFILE *cfp);

static int check_add_if_child (CffDirLNode *ent,
                               const char *child_name,
                               CffDirLNode *parent,
                               LNode *base,
                               COMPOUNDFILE *cfp);

static void make_storage_file(CffDirLNode *dir,
                              LNode *base,
                              COMPOUNDFILE *cfp,
                              CffStreamFilter filter,
                              int no_overwrite,
                              const char *path);


static void collect_sibling_name(LNode *b, DirID_t id, COMPOUNDFILE *cfp)
{
    LNode *ret;
    char *name;
    if (id == CFF_NO_DIRID)
        return;
    name = strdup_from_utf16(cfp->dir[id].name, 32);
    if (name == NULL)
        return;
    ret = llist_prepend_copy(b, name, strlen(name) + 1);
    free(name);
    if (ret == NULL)
        return;
    collect_sibling_name(b, cfp->dir[id].left_dirid, cfp);
    collect_sibling_name(b, cfp->dir[id].right_dirid, cfp);
}

static LNode *make_direct_child_name_list(const char *storage_path,
                                          COMPOUNDFILE *cfp)
{
    LNode *base;
    DirID_t id;
    DirID_t cid;
    if (storage_path == NULL)
        return NULL;
    id = cff_get_dirid(storage_path, cfp);
    if (CFF_NO_DIRID == id)
        return NULL;
    cid = cfp->dir[id].child_dirid;
    if (CFF_NO_DIRID == cid)
        return NULL;
    base = llist_new_base();
    if (base == NULL)
        return NULL;
    collect_sibling_name(base, cid, cfp);
    return base;
}

static char *make_new_path(const char *prefix, const char *append)
{
    char *n;
    const char *a;
    const char *p;
    int p_len;
    int a_len;
    int n_len;
    char *slash;

    p = prefix;
    if (append[0] == '/')
        a = append + 1;
    else
        a = append;

    p_len = strlen(p);
    a_len = strlen(a);

    if (p[p_len - 1] == '/')
        slash = "";
    else
        slash = "/";

    n_len = p_len + strlen(slash) + a_len;
    n = malloc(n_len + 1);
    if (n == NULL)
        return NULL;
    snprintf(n, n_len + 1, "%s%s%s", p, slash, a);
    return n;
}

LNode *cff_make_dir_list(COMPOUNDFILE *cfp)
{
    unsigned int total;
    int i;
    CffDirLNode ent;
    CffDirLNode *root_entry;
    LNode *n;
    LNode *base;

    base = llist_new_base();
    if (base == NULL)
        return NULL;

    total = CFF_DIR_ENTRY_NUM(cfp); 
    for (i = total - 1; i >= 0; --i) {
        ent.id = i;
        ent.name = strdup_from_utf16(cfp->dir[i].name, 32);
        ent.type = cfp->dir[i].object_type;
        ent.path = NULL;
        ent.parent_id = CFF_NO_DIRID;
        n = llist_prepend_copy(base, &ent, sizeof(ent));
        if (n == NULL || ent.name == NULL)
            goto error_handler;
    }

    root_entry = base->next->data;
    root_entry->path = strdup("/");
    if (root_entry->path == NULL)
        goto error_handler;

    add_path_parent(root_entry, base, cfp);
    return base;

error_handler:
    llist_destroy(base);
    return NULL;
}

void cff_destroy_dir_list(LNode *base)
{
    if (base == NULL)
        return;
    llist_destroy_with_function(base, (LlistDestroyF) free_entry);
}

static void free_entry(CffDirLNode *ent)
{
    if (ent == NULL)
        return;
    free(ent->name);
    free(ent->path);
    free(ent);
}

static void add_path_parent(CffDirLNode *parent, LNode *base, COMPOUNDFILE *cfp)
{
    int ret;
    LNode *n;
    LNode *cn;
    CffDirLNode *ent;
    char *child_name;
    LNode *child_name_list;

    if (parent == NULL || base == NULL)
        return;

    child_name_list = make_direct_child_name_list(parent->path, cfp);
    if (child_name_list == NULL)
        return;

    for (n = base->next; n != NULL; n = n->next) {
        ent = n->data;
        if (ent == NULL)
            goto finish;
        for (cn = child_name_list->next; cn != NULL; cn = cn->next) {
            child_name = cn->data;
            if (child_name == NULL)
                goto finish;
            ret = check_add_if_child(ent, child_name,
                                     parent, base, cfp);
            if (-1 == ret)
                goto finish;
        }
    }
finish:
    llist_destroy(child_name_list);
}


static int check_add_if_child (CffDirLNode *ent,
                               const char *child_name,
                               CffDirLNode *parent,
                               LNode *base,
                               COMPOUNDFILE *cfp)
{
    if (ent == NULL || child_name == NULL || parent == NULL)
        return -1;
    if (0 == strcmp(ent->name, child_name)) {
        ent->parent_id = parent->id;
        ent->path = make_new_path(parent->path, child_name);
        if (ent->path == NULL)
            return -1;
        if (ent->type == CFF_STORAGE_DIR)
            add_path_parent(ent, base, cfp);
    }
    return 0;
}

CffDirLNode *cff_find_dir_node_by_id(unsigned int id, LNode *base)
{
    CffDirLNode *ret;
    LNode *n;
    if (base == NULL)
        return NULL;
    for (n = base->next; n != NULL; n = n->next) {
        ret = n->data;
        if (id == ret->id)
            return ret;
    }
    return NULL;
}

CffDirLNode *cff_find_dir_node(const char *name, LNode *base)
{
    CffDirLNode *ret;
    LNode *n;
    if (base == NULL || name == NULL)
        return NULL;
    for (n = base->next; n != NULL; n = n->next) {
        ret = n->data;
        if (ret->name == NULL)
            continue;
        if (0 == strcmp(name, ret->name))
            return ret;
    }
    return NULL;
}

static unsigned count_digit(unsigned n)
{
    unsigned c = 1;
    while (n /= 10)
        ++c;
    return c;
}

static char *give_default_name(void)
{
    static char *default_name = "none";
    static int default_num = 0;
    char *str;
    int size;

    size = strlen(default_name) + count_digit(default_num) + 1;
    str = malloc(size);
    if (str == NULL)
        return NULL;

    snprintf(str, size, "%s%d", default_name, default_num);
    ++default_num;
    return str;
}

static char *make_split_filename(const char *path, const char *filename)
{
    static char *no = "`!@#$%^&*|\\?/:\005";

    char *p;
    char *str;

    if (path == NULL || filename == NULL)
        return NULL;

    str = make_filename(path, filename);
    if (str == NULL)
        return NULL;

    for (p = no; *p != '\0'; ++p)
        del_chr_in_str(str, *p);

    if (str[0] != '\0')
        return str;
    free(str);
    str = give_default_name();
    return str;
}

static void make_stream_file(CffDirLNode *d,
                             COMPOUNDFILE *c,
                             CffStreamFilter filter,
                             int no_overwrite,
                             const char *path)
{
    char *filename;
    CFFSTREAM *s = NULL;
    CFFSTREAM *temp;
    FILE *fp = NULL;

    filename = make_split_filename(path, d->name);
    if (filename == NULL)
        goto free_all;

    s = cff_salloc_by_id(d->id, c);
    if (s == NULL)
        goto free_all;

    if (filter != NULL) {
        temp = filter(s);
        if (temp == NULL)
            goto free_all;
        cff_free(s);
        s = temp;
    }

    if (0 == check_file(filename, no_overwrite))
        fp = fopen(filename, "wb");

    if (fp == NULL)
        goto free_all;

    if (1 != fwrite(s->data, s->data_size, 1, fp))
        fprintf(stderr, "Cannot fully write %s\n", filename);

free_all:
    if (fp != NULL)
        fclose(fp);
    cff_free(s);
    free(filename);
}

void cff_make_split_stream_storage(CffDirLNode *node,
                                   LNode *base,
                                   COMPOUNDFILE *cfp,
                                   CffStreamFilter filter,
                                   int no_overwrite,
                                   const char *path)
{
    if (node == NULL)
        return;
    if (node->name == NULL || node->name[0] == '\0')
        return;

    switch (node->type) {
    case CFF_STREAM_DIR:
        make_stream_file(node, cfp, filter, no_overwrite, path);
        break;
    case CFF_ROOT_DIR:
    case CFF_STORAGE_DIR:
        make_storage_file(node, base, cfp, filter, no_overwrite, path);
        break;
    default:
        break;
    }
}

static void make_storage_file(CffDirLNode *dir,
                              LNode *base,
                              COMPOUNDFILE *cfp,
                              CffStreamFilter filter,
                              int no_overwrite,
                              const char *path)
{
    int err;
    char *start_dir;
    CffDirLNode *d;
    LNode *n;

    if (dir == NULL || base == NULL)
        return;

    start_dir = strdup_pwd();
    if (start_dir == NULL)
        return;

    if (-1 == make_dir_chdir_with_suffix(path, dir->name))
        return;

    for (n = base->next; n != NULL; n = n->next) {
        d = n->data;
        if (d == NULL || d->name == NULL)
            break;
        if (d->parent_id == dir->id)
            cff_make_split_stream_storage(d, base, cfp,
                                          filter, no_overwrite, path);
    }

    if (-1 == chdir(start_dir)) {
        err = errno;
        file_err_msg(err, start_dir);
    }
    free(start_dir);
}
