#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "fileutil.h"

char *strdup_basename(const char *path)
{
    char *final_delim;
#ifdef IN_WINDOWS
    char delim = '\\';
#else
    char delim = '/';
#endif
    if (path == NULL)
        return NULL;

    final_delim = strrchr(path, delim);

    if (final_delim != NULL)
        return strdup(final_delim + 1);
    else
        return strdup(path);
}

/* ext: path의 확장자. 점 포함. 빈 문자열도 가능하다.  */
static char *make_filename_with_ext(const char *path,
                                    const char *ext,
                                    const char *suffix)
{
    char *ext_pos;
    unsigned new_len;
    char *new_filename;
    unsigned append_len;
    char *append;
    char *basename;
    char *fmt;

    if (path == NULL || ext == NULL || suffix == NULL)
        return NULL;

    new_filename = NULL;
    append_len = strlen(ext) + strlen(suffix) + 1;

    append = malloc(append_len);
    basename = strdup_basename(path);
    if (basename == NULL || append == NULL)
        goto new_filename_error;

    new_len = strlen(basename) + append_len;
    new_filename = malloc(new_len);
    if (new_filename == NULL)
        goto new_filename_error;

    strcpy(new_filename, basename);
    free(basename);

    if (*ext != '\0') {
        ext_pos = strstr(new_filename, ext);
        if (ext_pos != NULL) {
            *ext_pos = '_';
            ext_pos[1] = '\0';
            ++ext;
        }
    }

    if (*new_filename == '\0') {
        fmt = "%s%s";
    } else {
        fmt = "%s_%s";
        ++append_len;
    }

    snprintf(append, append_len, fmt, ext, suffix);
    strcat(new_filename, append);

    free(append);
    return new_filename;

new_filename_error:
    free(new_filename);
    free(append);
    free(basename);
    return NULL;
}

/* 파일 확장자를 스스로 검사한다.  반환 문자열은 다 쓰고 free해야 한다.  */
char *make_filename(const char *path, const char *suffix)
{
    char *ext;
    char *basename;
    char *new_filename;

    if (path == NULL || suffix == NULL)
        return NULL;

    basename = strdup_basename(path);
    if (basename == NULL)
        return NULL;

    ext = strrchr(basename, '.');
    if (ext == NULL)
        ext = "";

    new_filename = make_filename_with_ext(basename, ext, suffix);

    free(basename);
    return new_filename;
}

/* stat으로 파일 종류를 알아낸다.  에러 발생시 -errno 반환한다.  */
enum _check_file {REGULAR_FILE, DIRECTORY, ANOTHER_FILE};
static int check_file_type(const char *path)
{
    struct stat st;
    int ret;

    if (path == NULL)
        return -EFAULT;

    ret = stat(path, &st);
    if (ret == -1)
        return -errno;

    if (S_ISDIR(st.st_mode))
        return DIRECTORY;
    if (S_ISREG(st.st_mode))
        return REGULAR_FILE;

    return ANOTHER_FILE;
}

int check_file(const char *path, int no_overwrite)
{
    int chk = check_file_type(path);
    if (chk >= 0) {
        if (no_overwrite) {
            fprintf(stderr, "%s already exists\n", path);
            return -1;
        }
    } else {
        int err = -chk;
        if (err != ENOENT) {
            file_err_msg(err, path);
            return -1;
        }
    }
    return 0;
}

void file_err_msg(int err, const char *path)
{
    switch (err) {
    case 0:
        break;
    case ENOENT:
        fprintf(stderr, "%s does not exist\n", path);
        break;
    case EEXIST:
        fprintf(stderr, "%s already exists\n", path);
        break;
    case ENOTDIR:
        fprintf(stderr, "%s is not a directory\n", path);
        break;
    case EPERM:
    case EACCES:
        fprintf(stderr, "Permission is denied: %s\n", path);
        break;
    case ENAMETOOLONG:
        fprintf(stderr, "Too long path: %s\n", path);
        break;
    case ENOMEM:
        fputs("Kernel wants more memory\n", stderr);
        break;
    case EROFS:
        fprintf(stderr, "%s refers to a directory on a read-only file system\n", path);
        break;
    case ENOSPC:
        fprintf(stderr, "No room for %s\n", path);
        break;
    case EFAULT:
        fputs("Wrong pointer or bad address\n", stderr);
        break;
    case EIO:
        fputs("An I/O error occurred\n", stderr);
        break;
    default:
        fprintf(stderr, "An error occurred with %s\n", path);
        break;
    }
}

int make_dir_chdir(const char *path)
{
    int err;
    int chk = check_file_type(path);

    switch (chk) {
    case DIRECTORY:
        break;
    case -ENOENT:
#ifdef IN_WINDOWS
        if (0 == mkdir(path))
#else
        if (0 == mkdir(path, 0700))
#endif
            break;
        err = errno;
        file_err_msg(err, path);
        return -1;
    case REGULAR_FILE:
    case ANOTHER_FILE:
        fprintf(stderr, "%s is not a directory\n", path);
        return -1;
    default:
        file_err_msg(-chk, path);
        return -1;
    }

    if (0 == chdir(path))
        return 0;
    err = errno;
    file_err_msg(err, path);
    return -1;
}

#define DIRBUFSIZE 512
char *strdup_pwd()
{
    int err;
    char *ret;
    char *dirbuf;
    static unsigned int nbuf = 1;

    while (1) {
        dirbuf = malloc(DIRBUFSIZE * nbuf);
        if (dirbuf == NULL)
            return NULL;
        ret = getcwd(dirbuf, DIRBUFSIZE);
        if (ret != NULL)
            break;
        err = errno;
        free(dirbuf);
        if (err == ERANGE)
            ++nbuf;
        else
            return NULL;
    }

    ret = strdup(dirbuf);
    free(dirbuf);
    return ret;
}

int redirect_stdout_to_file(int no_overwrite,
                            const char *path,
                            const char *suffix)
{
    FILE *fp;
    char *new_filename;

    new_filename = make_filename(path, suffix);
    if (new_filename == NULL) {
        fprintf(stderr, "Cannot make a new filename of %s\n", path);
        return -1;
    }

    if (-1 == check_file(new_filename, no_overwrite)) {
        free(new_filename);
        return -1;
    }

    fp = freopen(new_filename, "w", stdout);
    if (fp == NULL) {
        fprintf(stderr, "Cannot create %s\n", new_filename);
        free(new_filename);
        return -1;
    }

    fprintf(stderr, "Created %s\n", new_filename);
    free(new_filename);
    return 0;
}

int make_dir_chdir_with_suffix(const char *path, const char *suffix)
{
    char *dirname;
    if (path == NULL || suffix == NULL)
        return -1;

    dirname = make_filename(path, suffix);
    if (dirname == NULL)
        return -1;

    make_dir_chdir(dirname);
    free(dirname);
    return 0;
}
