/* hwp5read.c

   hwp5dump는 한글과컴퓨터의 한글 문서 파일 구조 정의
   공개 정책을 따르는 2차 저작물입니다.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>
#include "cffread.h"
#include "unicode_locale.h"
#include "unicode_string.h"
#include "hwp5read.h"
#include "hwp5type.h"

#define struct_major_ver(v) ((unsigned)((v) & 0xFF000000) >> 24)
#define struct_minor_ver(v) ((unsigned)((v) & 0x00FF0000) >> 16)
#define record_major_ver(v) ((unsigned)((v) & 0x0000FF00) >> 8)
#define record_minor_ver(v) ((v) & 0x000000FF)

int   hwp5_get_header    (HWP5FILE *hfp);

int   hwp5_alloc_summary (HWP5FILE *hfp);
void  hwp5_free_summary  (HWP5FILE *hfp);

int   hwp5_alloc_prvtext (HWP5FILE *hfp);
void  hwp5_free_prvtext  (HWP5FILE *hfp);


HWP5FILE *hwp5_open(const char *path, const char *mode)
{
    HWP5FILE *hfp = malloc(sizeof(HWP5FILE));
    if (hfp == NULL)
        return NULL;
    memset(hfp, 0, sizeof(HWP5FILE));

    hfp->cfp = cff_open(path, mode);
    if (hfp->cfp == NULL) {
        free(hfp);
        return NULL;
    }

    if (hwp5_get_header(hfp) == -1) {
        cff_close(hfp->cfp);
        free(hfp);
        return NULL;
    }

    /* 미리보기나 요약이 없어도 읽어본다.  */
    if (hwp5_alloc_prvtext(hfp) == -1)
        fputs("Cannot get preview text\n", stdout);
    if (hwp5_alloc_summary(hfp) == -1)
        fputs("Cannot get summary infomation\n", stdout);

    return hfp;
}


void hwp5_close(HWP5FILE *hfp)
{
    if (hfp == NULL)
        return;
    hwp5_free_prvtext(hfp);
    hwp5_free_summary(hfp);
    cff_close(hfp->cfp);
    free(hfp);
}

int hwp5_get_header(HWP5FILE *hfp)
{
    CFFSTREAM *stream = cff_salloc("FileHeader", hfp->cfp);
    if (stream == NULL)
        return -1;
    hfp->header = *(HWPHEADER *)stream->data;
    cff_free(stream);
    if (memcmp("HWP Document File", hfp->header.signature, 18) != 0)
        return -1;
    if ((hfp->header.version & 0x05000000) != 0x05000000)
        return -1;
    /* 패스워드 있는 문서도 한번 읽어본다.  */
    if (HWP_IS_LOCKED(hfp))
        fputs("This hwp file is locked\n", stdout);

    return 0;
}


void hwp5_get_version_string(char *str, size_t str_size, HWP5FILE *hfp)
{
    snprintf(str, str_size, "%d.%d.%d.%d",
                           struct_major_ver(hfp->header.version),
                           struct_minor_ver(hfp->header.version),
                           record_major_ver(hfp->header.version),
                           record_minor_ver(hfp->header.version));
}


/* \005HwpSummaryInformation에서 6개의 데이터만을 뽑아낸다.
   이 데이터들은 모두 UTF-16 문자열인데 UTF-8 또는 C locale로 바꿔서 읽는다.
   Property ID들을 확인하지 않고 저장된 순서에만 의존하고 있다. */
int hwp5_alloc_summary(HWP5FILE *hfp)
{
    DWORD property_set_size;
    BYTE *ptr;

    DWORD offset; /* 기준은 스트림이 아니라 Property Set 시작 위치 */
    DWORD num;    /* UTF-16 문자열의 널 포함 WCHAR 배열 요소 개수 */
    WCHAR *temp_str;

    DWORD i = 0;
    DWORD j = 0;
    char *str[6] = {0, 0, 0, 0, 0, 0};

    CFFSTREAM *stream = cff_salloc("\005HwpSummaryInformation", hfp->cfp);
    if (stream == NULL)
        return -1;

    /* ptr을 Property Set 시작 위치로 맞춘다. */
    ptr = (BYTE *)stream->data + 48;
    property_set_size = *(DWORD *)ptr;
    if (stream->data_size != property_set_size + 48)
        goto error_handler;

    for (i = 0; i < 6; i++) {
        offset = *(DWORD *)(ptr + 12 + 8*i);
        num = *(DWORD *)(ptr + offset + 4);
        if (offset + num * sizeof(WCHAR) > property_set_size ||
            0x1F != *(WORD *)(ptr + offset))
            goto error_handler;
        temp_str = (WCHAR *)(ptr + offset + 8);
        str[i] = strdup_from_utf16(temp_str, num);
        if (str[i] == NULL)
            goto error_handler;
    }

    hfp->summary.title   = str[0];
    hfp->summary.subject = str[1];
    hfp->summary.author  = str[2];
    hfp->summary.date    = str[3];
    hfp->summary.keyword = str[4];
    hfp->summary.others  = str[5];

    cff_free(stream);
    return 0;

error_handler:
    for (j = 0; j < i; j++)
        free(str[j]);
    cff_free(stream);
    return -1;
}

void hwp5_free_summary(HWP5FILE *hfp)
{
    if (hfp == NULL)
        return;
    free(hfp->summary.title);
    free(hfp->summary.subject);
    free(hfp->summary.author);
    free(hfp->summary.date);
    free(hfp->summary.keyword);
    free(hfp->summary.others);
}

/* PrvText는 끝에 널 문자가 없고 인코딩이 UTF-16이다.
   리눅스에서 편하게 읽으려면 메모리로 읽어 들일 때
   이것에 널 문자를 붙이고 인코딩을 UTF-8로 바꿔야 한다.
   그리고 Carriage return 문자(0x0D)는 제거한다. */
int hwp5_alloc_prvtext(HWP5FILE *hfp)
{
    DWORD utf16_num;
    char *prvtext;

    CFFSTREAM *stream = cff_salloc("PrvText", hfp->cfp);
    if (stream == NULL)
        return -1;

    utf16_num = stream->data_size / sizeof(WCHAR);
    prvtext = strdup_from_utf16(stream->data, utf16_num);
    if (prvtext == NULL) {
        cff_free(stream);
        return -1;
    }
    del_chr_in_str(prvtext, 0x0D);
    hfp->prvtext = prvtext;
    cff_free(stream);
    return 0;
}

void hwp5_free_prvtext(HWP5FILE *hfp)
{
    if (hfp == NULL)
        return;
    free(hfp->prvtext);
}

#define BUFFER_MORE 500000
#define SUSPICIOUS_SIZE 50000000
CFFSTREAM *hwp5_decompress(const CFFSTREAM *stream)
{
    CFFSTREAM *decom_stream;
    void *decom_data;
    uLongf decom_size;

    z_stream zs;
    int z_ret;

    if (stream == NULL)
        return NULL;

    decom_size = *(uint32_t *)(stream->data + stream->data_size - 4);
    if (decom_size == 0 || decom_size > SUSPICIOUS_SIZE)
        decom_size = (decom_size + SUSPICIOUS_SIZE) / 1000;

    decom_data = malloc(decom_size);
    if (decom_data == NULL)
        return NULL;

    zs.next_in = stream->data;
    zs.avail_in = stream->data_size - 4;
    zs.next_out = decom_data;
    zs.avail_out = decom_size;
    zs.zalloc = Z_NULL;
    zs.zfree = Z_NULL;
    zs.opaque = Z_NULL;

    if (inflateInit2(&zs, -15) != Z_OK)
        goto error_handler;

    while (1) {
        z_ret = inflate(&zs, Z_FINISH);
        if (z_ret == Z_STREAM_END)
            break;
        /* 파일에 저장돼 있는 압축전크기가 틀릴 수도 있다.  */
        if (z_ret == Z_BUF_ERROR) {
            void *temp;
            if (zs.avail_in == 0 && zs.avail_out > 0) {
                decom_size -= zs.avail_out;
                temp = realloc(decom_data, decom_size);
                if (temp == NULL)
                    goto error_handler;
                decom_data = temp;
                zs.avail_out = 0;
                break;
            } else if (zs.avail_in > 0 && zs.avail_out == 0) {
                uLongf new_size = decom_size + BUFFER_MORE;
                temp = realloc(decom_data, new_size);
                if (temp == NULL)
                    goto error_handler;
                decom_data = temp;
                zs.avail_out = BUFFER_MORE;
                zs.next_out = (unsigned char *) decom_data + decom_size;
                decom_size = new_size;
                continue;
            } else {
                goto error_handler;
            }
        } else {
            goto error_handler;
        }
    }

    if (inflateEnd(&zs) != Z_OK)
        goto error_handler;

    decom_stream = malloc(sizeof(CFFSTREAM));
    if (decom_stream == NULL)
        goto error_handler;

    decom_stream->data = decom_data;
    decom_stream->data_size = decom_size;
    decom_stream->alloc_size = decom_size;

    return decom_stream;

error_handler:
    free(decom_data);
    return NULL;
}


HDRSTREAM *hwp5_alloc_hdrstream(const char *stream_path, HWP5FILE *hfp)
{
    CFFSTREAM *cs, *cs_temp;
    HDRSTREAM *hs;

    cs = cff_salloc(stream_path, hfp->cfp);
    if (cs == NULL)
        return NULL;

    if (HWP_IS_COMPRESSED(hfp)) {
        cs_temp = hwp5_decompress(cs);
        cff_free(cs);
        if (cs_temp == NULL)
            return NULL;
        cs = cs_temp;
    }

    hs = malloc(sizeof(HDRSTREAM));
    if (hs == NULL) {
        cff_free(cs);
        return NULL;
    }

    hs->stream = cs;
    hs->offset = 0;

    return hs;
}


void hwp5_free_hdrstream(HDRSTREAM *hdrstream)
{
    if (hdrstream == NULL)
        return;
    cff_free(hdrstream->stream);
    free(hdrstream);
}


/* tag_id가 HDR_NEXT_RECORD이면 바로 다음 레코드를 반환함 */
HDR *hwp5_alloc_hdr(int tag_id, HDRSTREAM *h)
{
    BYTE *ptr, *stream_start;
    HDR *hdr;

    if (h->offset >= h->stream->data_size)
        return NULL;

    hdr = malloc(sizeof(HDR));
    if (hdr == NULL)
        return NULL;

    memset(hdr, 0, sizeof(HDR));
    ptr = stream_start = h->stream->data;
    ptr += h->offset;

    do {
        ptr += hdr->size;
        memcpy(&(hdr->header), ptr, sizeof(HDRHEADER));
        ptr += sizeof(HDRHEADER);

        if (hdr->header.size == 0xFFF) {
            memcpy(&(hdr->size), ptr, sizeof(DWORD));
            ptr += sizeof(DWORD);
        } else {
            hdr->size = hdr->header.size;
        }

        if (hdr->size + ptr - stream_start > h->stream->data_size) {
            free(hdr);
            return NULL;
        }

    } while (tag_id != HDR_NEXT_RECORD && hdr->header.tag_id != tag_id);

    hdr->data = malloc(hdr->size);
    if (hdr->data == NULL) {
        free(hdr);
        return NULL;
    }
    memcpy(hdr->data, ptr, hdr->size);
    h->offset = hdr->size + ptr - stream_start;

    return hdr;
}


void hwp5_free_hdr(HDR *hdr)
{
    if (hdr == NULL)
        return;
    free(hdr->data);
    free(hdr);
}
